div{hoo:3/4;goo:0.75;foo:20;foo:1;coo:0.88889;bar:true;bar:false and true;bar:false;blah:a;blah:0.75;blah:b;bloo:0.75;blix:"hey 0.75 ho"}@media screen and (hux: 3 / 4){div{color:red}}div{blah:"ho 0.75 ho"}span{fludge:0.75}
