a b { color: red; }
a b :first-child, a b :nth-of-type(-2n+1) { blah: bloo; }
a b :first-child .foo#bar:nth-child(even), a b :nth-of-type(-2n+1) .foo#bar:nth-child(even) { hoo: goo; }
a b :first-child ::after, a b :nth-of-type(-2n+1) ::after { content: "glux"; color: green; }
a b :first-child :not(.foo), a b :nth-of-type(-2n+1) :not(.foo) { hoo: boo; }
a b :first-child , a b :nth-of-type(-2n+1)  { hoo: boo; }
a b :first-child a, a b :nth-of-type(-2n+1) a { b: c; }
